/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.51 from the
 * contents of test.xs. Do not edit this file, edit test.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "t/test.xs"
/*  You may distribute under the terms of either the GNU General Public License
 *  or the Artistic License (the same terms as Perl itself)
 *
 *  (C) Paul Evans, 2024 -- leonerd@leonerd.org.uk
 */

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "DataChecks.h"

#define HAVE_PERL_VERSION(R, V, S) \
    (PERL_REVISION > (R) || (PERL_REVISION == (R) && (PERL_VERSION > (V) || (PERL_VERSION == (V) && (PERL_SUBVERSION >= (S))))))

#include "optree-additions.c.inc"

#line 28 "t/test.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

#line 172 "t/test.c"

XS_EUPXS(XS_t__test_make_checkdata); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_t__test_make_checkdata)
{
    dVAR; dXSARGS;
    if (items < 2 || items > 3)
       croak_xs_usage(cv,  "checkspec, name, constraint= &PL_sv_undef");
    {
	struct DataChecks_Checker *	RETVAL;
	dXSTARG;
	SV *	checkspec = ST(0)
;
	SV *	name = ST(1)
;
	SV *	constraint;

	if (items < 3)
	    constraint = &PL_sv_undef;
	else {
	    constraint = ST(2)
;
	}
#line 26 "t/test.xs"
    RETVAL = make_checkdata(checkspec);
    gen_assertmess(RETVAL, name, constraint);
#line 198 "t/test.c"
	XSprePUSH;
	PUSHi(PTR2IV(RETVAL));
    }
    XSRETURN(1);
}


XS_EUPXS(XS_t__test_free_checkdata); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_t__test_free_checkdata)
{
    dVAR; dXSARGS;
    if (items != 1)
       croak_xs_usage(cv,  "checker");
    {
	struct DataChecks_Checker *	checker = INT2PTR(struct DataChecks_Checker *,SvIV(ST(0)))
;

	free_checkdata(checker);
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_t__test_check_value); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_t__test_check_value)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "checker, value");
    {
	bool	RETVAL;
	struct DataChecks_Checker *	checker = INT2PTR(struct DataChecks_Checker *,SvIV(ST(0)))
;
	SV *	value = ST(1)
;

	RETVAL = check_value(checker, value);
	ST(0) = boolSV(RETVAL);
    }
    XSRETURN(1);
}


XS_EUPXS(XS_t__test_assert_value); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_t__test_assert_value)
{
    dVAR; dXSARGS;
    if (items != 2)
       croak_xs_usage(cv,  "checker, value");
    {
	struct DataChecks_Checker *	checker = INT2PTR(struct DataChecks_Checker *,SvIV(ST(0)))
;
	SV *	value = ST(1)
;

	assert_value(checker, value);
    }
    XSRETURN_EMPTY;
}


XS_EUPXS(XS_t__test_make_asserter_sub); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_t__test_make_asserter_sub)
{
    dVAR; dXSARGS;
    if (items < 1 || items > 2)
       croak_xs_usage(cv,  "checker, flagname= &PL_sv_undef");
    {
	SV *	RETVAL;
	struct DataChecks_Checker *	checker = INT2PTR(struct DataChecks_Checker *,SvIV(ST(0)))
;
	SV *	flagname;

	if (items < 2)
	    flagname = &PL_sv_undef;
	else {
	    flagname = ST(1)
;
	}
#line 39 "t/test.xs"
  {
    if(!PL_parser) {
      /* We need to generate just enough of a PL_parser to keep newSTATEOP()
       * happy, otherwise it will SIGSEGV
       */
      SAVEVPTR(PL_parser);
      Newxz(PL_parser, 1, yy_parser);
      SAVEFREEPV(PL_parser);

      PL_parser->copline = NOLINE;
      PL_parser->preambling = NOLINE;
    }

    U32 flags = 0;
    if(flagname && SvOK(flagname)) {
      if(SvPOK(flagname) && strEQ(SvPVX(flagname), "void"))
        flags = OPf_WANT_VOID;
    }

    I32 floorix = start_subparse(FALSE, 0);
    OP *body = newLISTOPn(OP_RETURN, 0,
      make_assertop_flags(checker, flags, newSLUGOP(0)),
      NULL);
    CV *cv = newATTRSUB(floorix, NULL, NULL, NULL, body);
    RETVAL = newRV_noinc((SV *)cv);
  }
#line 305 "t/test.c"
	RETVAL = sv_2mortal(RETVAL);
	ST(0) = RETVAL;
    }
    XSRETURN(1);
}

#ifdef __cplusplus
extern "C" {
#endif
XS_EXTERNAL(boot_t__test); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_t__test)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        newXS_deffile("t::test::make_checkdata", XS_t__test_make_checkdata);
        newXS_deffile("t::test::free_checkdata", XS_t__test_free_checkdata);
        newXS_deffile("t::test::check_value", XS_t__test_check_value);
        newXS_deffile("t::test::assert_value", XS_t__test_assert_value);
        newXS_deffile("t::test::make_asserter_sub", XS_t__test_make_asserter_sub);

    /* Initialisation Section */

#line 69 "t/test.xs"
  boot_data_checks(0);

#line 351 "t/test.c"

    /* End of Initialisation Section */

#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

#ifdef __cplusplus
}
#endif
